@@ -1,26 +1,33 @@
+/* -*- mode: C; indent-tabs-mode: t; tab-width: 8; c-basic-offset: 2; -*- */
+
 /*
  * This file is part of Seed, the GObject Introspection<->Javascript bindings.
  *
- * Seed is free software: you can redistribute it and/or modify 
+ * Seed is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version. 
- * Seed is distributed in the hope that it will be useful, 
- * but WITHOUT ANY WARRANTY; without even the implied warranty of 
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
- * GNU Lesser General Public License for more details. 
- * You should have received a copy of the GNU Lesser General Public License 
- * along with Seed.  If not, see <http://www.gnu.org/licenses/>. 
+ * the License, or (at your option) any later version.
+ * Seed is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU Lesser General Public License for more details.
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with Seed.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Copyright (C) Robert Carr 2008 <carrr@rpi.edu>
+ * Copyright (C) Robert Carr 2009 <carrr@rpi.edu>
  */
 
 #include "seed-private.h"
+#include "seed-path.h"
 
 #include <sys/types.h>
 #include <sys/wait.h>
-#include <string.h>
 #include <stdarg.h>
+#include <string.h>
+
+#include "config.h"
+
+JSObjectRef function_proto;
 
 JSObjectRef seed_obj_ref;
 
@@ -40,6 +47,10 @@ GIBaseInfo *base_info_info = 0;
 GQuark js_ref_quark;
 
 guint seed_debug_flags = 0;	/* global seed debug flag */
+gboolean seed_arg_print_version = FALSE; // Flag to print version and quit
+
+__thread JSObjectRef seed_next_gobject_wrapper = NULL;
+
 
 #ifdef SEED_ENABLE_DEBUG
 static const GDebugKey seed_debug_keys[] = {
@@ -51,18 +62,49 @@ static const GDebugKey seed_debug_keys[] = {
   {"structs", SEED_DEBUG_STRUCTS},
   {"construction", SEED_DEBUG_CONSTRUCTION},
   {"gtype", SEED_DEBUG_GTYPE},
-  {"importer", SEED_DEBUG_IMPORTER}
+  {"importer", SEED_DEBUG_IMPORTER},
+  {"module", SEED_DEBUG_MODULE}
 };
 #endif /* SEED_ENABLE_DEBUG */
 
+static bool 
+seed_gobject_has_instance (JSContextRef ctx, JSObjectRef constructor,
+			  JSValueRef possible_instance, JSValueRef* exception)
+{
+  GType constructor_type, value_type;
+  if (JSValueIsNull (ctx, possible_instance) ||
+      !JSValueIsObject (ctx, possible_instance) ||
+      !JSValueIsObjectOfClass (ctx, possible_instance, gobject_class))
+    return FALSE;
+  
+  constructor_type = (GType) JSObjectGetPrivate (constructor);
+  value_type = G_OBJECT_TYPE ((GObject *)
+			      JSObjectGetPrivate ((JSObjectRef) possible_instance));
+
+  return g_type_is_a (value_type, constructor_type);
+}
+
+
+/**
+ * seed_prepare_global_context:
+ * @ctx: A #SeedContext on which to add the default set of global objects.
+ *
+ * Adds the default set of global objects (imports, GType, Seed, and print)
+ * to a fresh #SeedContext.
+ *
+ */
 void
 seed_prepare_global_context (JSContextRef ctx)
 {
   JSObjectRef global = JSContextGetGlobalObject (ctx);
-  
+
   seed_object_set_property (ctx, global, "imports", importer);
   seed_object_set_property (ctx, global, "GType", seed_gtype_constructor);
   seed_object_set_property (ctx, global, "Seed", seed_obj_ref);
+  seed_object_set_property (ctx, global, "print", seed_print_ref);
+
+
+  JSEvaluateScript (ctx, defaults_script, NULL, NULL, 0, NULL);
 }
 
 static JSObjectRef
@@ -80,7 +122,7 @@ seed_struct_constructor_invoked (JSContextRef ctx,
     {
       if (!JSValueIsObject (ctx, arguments[0]))
 	{
-	  seed_make_exception (ctx, exception, "ArgmuentError",
+	  seed_make_exception (ctx, exception, "ArgumentError",
 			       "Constructor expects object as argument");
 	  return (JSObjectRef) JSValueMakeNull (ctx);
 	}
@@ -103,10 +145,9 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
   GParameter *params;
   GObjectClass *oclass;
   GObject *gobject;
-  // Do we free GParamSpecs...? It's not clear.
   GParamSpec *param_spec;
   gchar *prop_name;
-  gint i, nparams = 0, length;
+  gint i, nparams = 0, length, ri;
   JSObjectRef ret;
   JSPropertyNameArrayRef jsprops = 0;
   JSStringRef jsprop_name;
@@ -120,7 +161,7 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
 
   if (argumentCount > 1)
     {
-      seed_make_exception (ctx, exception, "ArgumentError", 
+      seed_make_exception (ctx, exception, "ArgumentError",
 			   "Constructor expects"
 			   " 1 argument, got %zd", argumentCount);
 
@@ -131,7 +172,7 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
     {
       if (!JSValueIsObject (ctx, arguments[0]))
 	{
-	  seed_make_exception (ctx, exception, "ArgmuentError",
+	  seed_make_exception (ctx, exception, "ArgumentError",
 			       "Constructor expects object as argument");
 	  g_type_class_unref (oclass);
 	  return (JSObjectRef) JSValueMakeNull (ctx);
@@ -140,12 +181,14 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
       jsprops = JSObjectCopyPropertyNames (ctx, (JSObjectRef) arguments[0]);
       nparams = JSPropertyNameArrayGetCount (jsprops);
     }
-  i = 0;
+  ri = i = 0;
 
   params = g_new0 (GParameter, nparams + 1);
   SEED_NOTE (INITIALIZATION, "Constructing object of type %s",
 	     g_type_name (type));
 
+  seed_next_gobject_wrapper = seed_make_wrapper_for_type (ctx, type);
+
   while (i < nparams)
     {
       GType type;
@@ -156,47 +199,51 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
       JSStringGetUTF8CString (jsprop_name, prop_name, length);
 
       param_spec = g_object_class_find_property (oclass, prop_name);
-      if (param_spec == NULL)
-	{
-	  seed_make_exception (ctx, exception, "PropertyError", 
-			       "Invalid property for construction: %s",
-			       prop_name);
 
-	  g_free (params);
-
-	  JSPropertyNameArrayRelease (jsprops);
-	  return (JSObjectRef) JSValueMakeNull (ctx);
-	}
-      // TODO: exception handling
       jsprop_value = JSObjectGetProperty (ctx,
 					  (JSObjectRef) arguments[0],
 					  jsprop_name, NULL);
 
+      if (param_spec == NULL)
+	{
+	  JSObjectSetProperty (ctx, seed_next_gobject_wrapper, jsprop_name,
+			       jsprop_value, 0, NULL);
+	  ++i;
+	  continue;
+	}
+      // TODO: exception handling
+
       if (g_type_is_a (param_spec->value_type, G_TYPE_ENUM))
 	type = G_TYPE_INT;
       else
 	type = param_spec->value_type;
 
       seed_gvalue_from_seed_value (ctx, jsprop_value,
-				   type, &params[i].value, exception);
+				   type, &params[ri].value, exception);
 
       if (*exception)
 	{
 	  g_free (params);
 	  JSPropertyNameArrayRelease (jsprops);
+	  seed_next_gobject_wrapper = NULL;
 	  return 0;
 	}
-      params[i].name = prop_name;
+      params[ri].name = prop_name;
 
       ++i;
+      ++ri;
     }
 
   if (jsprops)
     JSPropertyNameArrayRelease (jsprops);
 
-  gobject = g_object_newv (type, nparams, params);
 
-  if (G_IS_INITIALLY_UNOWNED (gobject))
+  gobject = g_object_newv (type, ri, params);
+
+
+  if (G_IS_INITIALLY_UNOWNED (gobject) && !g_object_is_floating (gobject))
+    g_object_ref (gobject);
+  else if (g_object_is_floating (gobject))
     g_object_ref_sink (gobject);
 
   if (!gobject)
@@ -204,13 +251,12 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
   else
     ret = (JSObjectRef) seed_value_from_object (ctx, gobject, exception);
 
-  for (i = 0; i < nparams; i++)
+  for (i = 0; i < ri; i++)
     {
       g_value_unset (&params[i].value);
     }
 
-  if (G_IS_INITIALLY_UNOWNED (gobject))
-    g_object_unref (gobject);
+  g_object_unref (gobject);
 
   g_type_class_unref (oclass);
 
@@ -220,29 +266,32 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
 }
 
 static JSValueRef
-seed_gobject_equals (JSContextRef ctx,
-		     JSObjectRef function,
-		     JSObjectRef this_object,
-		     size_t argumentCount,
-		     const JSValueRef arguments[], JSValueRef * exception)
+seed_gobject_property_type (JSContextRef ctx,
+			    JSObjectRef function,
+			    JSObjectRef this_object,
+			    size_t argumentCount,
+			    const JSValueRef arguments[],
+			    JSValueRef * exception)
 {
-  GObject *this, *that;
+  GParamSpec *spec;
+  gchar *name;
+  GObject *this;
 
   if (argumentCount != 1)
     {
-      seed_make_exception (ctx, exception, "ArgumentError", 
-			   "GObject equals comparison expected"
-			   " 1 argument, got %zd", argumentCount);
-
+      seed_make_exception (ctx, exception, "ArgumentError",
+			   "__property_type expects 1 argument"
+			   "got %zd", argumentCount);
       return JSValueMakeNull (ctx);
     }
 
-  this = seed_value_to_object (ctx, (JSValueRef) this_object, exception);
-  that = seed_value_to_object (ctx, arguments[0], exception);
+  this = seed_value_to_object (ctx, this_object, exception);
+  name = seed_value_to_string (ctx, arguments[0], exception);
 
-  if (this == that)
-    return seed_value_from_boolean (ctx, 1, 0);
-  return seed_value_from_boolean (ctx, 0, 0);
+  spec = g_object_class_find_property (G_OBJECT_GET_CLASS (this), name);
+  g_free (name);
+
+  return seed_value_from_long (ctx, spec->value_type, exception);
 }
 
 static JSValueRef
@@ -267,7 +316,36 @@ seed_gobject_method_finalize (JSObjectRef method)
     g_base_info_unref (info);
 }
 
-typedef void (*InitMethodCallback) (gint *argc, gchar ***argv);
+typedef void (*InitMethodCallback) (gint * argc, gchar *** argv);
+
+static gboolean
+seed_gobject_init_build_argv (JSContextRef ctx,
+			      JSObjectRef array,
+			      SeedArgvPrivates * priv, JSValueRef * exception)
+{
+  guint i, length;
+  JSValueRef jsl;
+
+  jsl = seed_object_get_property (ctx, array, "length");
+  if (JSValueIsNull (ctx, jsl) || JSValueIsUndefined (ctx, jsl))
+    return FALSE;
+
+  length = seed_value_to_uint (ctx, jsl, exception);
+  priv->argv = g_new (gchar *, length);
+  priv->argc = length;
+
+  for (i = 0; i < length; i++)
+    {
+      priv->argv[i] = seed_value_to_string (ctx,
+					    JSObjectGetPropertyAtIndex (ctx,
+									array,
+									i,
+									exception),
+					    exception);
+    }
+  return TRUE;
+
+}
 
 static JSValueRef
 seed_gobject_init_method_invoked (JSContextRef ctx,
@@ -280,39 +358,66 @@ seed_gobject_init_method_invoked (JSContextRef ctx,
   GIBaseInfo *info;
   GTypelib *typelib;
   InitMethodCallback c;
-  SeedArgvPrivates *priv;
+  SeedArgvPrivates *priv = NULL;
+  gboolean allocated = FALSE;
+
   if (argumentCount != 1 && argumentCount != 2)
     {
-      seed_make_exception (ctx, exception, 
-			   "ArgumentError", "init method expects 1 argument, got %zd", 
+      seed_make_exception (ctx, exception,
+			   "ArgumentError",
+			   "init method expects 1 argument, got %zd",
 			   argumentCount);
       return JSValueMakeUndefined (ctx);
     }
 
-  // TODO: In future support all arrays.Through a seperate code path.
-  if (!argumentCount == 2)
+  if (argumentCount == 1)
     {
-      if (JSValueIsNull (ctx, arguments[0]) || !JSValueIsObject (ctx, arguments[0]) ||
-	  !JSValueIsObjectOfClass (ctx, arguments[0], seed_argv_class))
+      if (JSValueIsNull (ctx, arguments[0])
+	  || !JSValueIsObject (ctx, arguments[0]))
+
 	{
 	  seed_make_exception (ctx, exception,
-			       "ArgumentError", "init method expects an argv object as argument");
+			       "ArgumentError",
+			       "init method expects an array object as argument");
+	  return JSValueMakeUndefined (ctx);
+	}
+      if (JSValueIsObjectOfClass (ctx, arguments[0], seed_argv_class))
+	{
+	  priv = JSObjectGetPrivate ((JSObjectRef) arguments[0]);
+	}
+      else
+	{
+	  priv = g_newa (SeedArgvPrivates, 1);
+	  if (!seed_gobject_init_build_argv (ctx,
+					     (JSObjectRef) arguments[0],
+					     priv, exception))
+	    {
+	      seed_make_exception (ctx, exception, "ArgumentError",
+				   "Init method expects an array as argument");
+	      return JSValueMakeUndefined (ctx);
+
+	    }
+	  allocated = TRUE;
 	}
     }
-  
+
   info = JSObjectGetPrivate (function);
   typelib = g_base_info_get_typelib (info);
-  g_typelib_symbol (typelib, g_function_info_get_symbol ((GIFunctionInfo *)info), (gpointer *)&c);
+  g_typelib_symbol (typelib,
+		    g_function_info_get_symbol ((GIFunctionInfo *) info),
+		    (gpointer *) & c);
   // Backwards compatibility
-  if (argumentCount == 2)
+  if (!priv)
     {
-      c(NULL, NULL);
+      c (NULL, NULL);
       return JSValueMakeUndefined (ctx);
     }
 
-  priv = JSObjectGetPrivate ((JSObjectRef)arguments[0]);
-  c(&priv->argc, &priv->argv);
-    
+  c (&priv->argc, &priv->argv);
+
+  if (allocated)
+    g_free (priv->argv);
+
   return JSValueMakeUndefined (ctx);
 }
 
@@ -332,6 +437,7 @@ seed_gobject_method_invoked (JSContextRef ctx,
   GArgument *out_args;
   GArgument *out_values;
   guint n_args, n_in_args, n_out_args, i;
+  guint in_args_pos, out_args_pos;
   GIArgInfo *arg_info;
   GITypeInfo *type_info;
   GIDirection dir;
@@ -340,8 +446,7 @@ seed_gobject_method_invoked (JSContextRef ctx,
 
   info = JSObjectGetPrivate (function);
   // We just want to check if there IS an object, not actually throw an
-  // exception if we don't
-  // get it.
+  // exception if we don't get it.
   if (!this_object || !
       ((object = seed_value_to_object (ctx, this_object, 0)) ||
        (object = seed_pointer_get_pointer (ctx, this_object))))
@@ -359,7 +464,9 @@ seed_gobject_method_invoked (JSContextRef ctx,
 
   for (i = 0; (i < (n_args)); i++)
     {
-      SEED_NOTE (INVOCATION, "Converting arg: %d of function %s, exception is %p", i, g_base_info_get_name (info), exception);
+      SEED_NOTE (INVOCATION,
+		 "Converting arg: %d of function %s, exception is %p", i,
+		 g_base_info_get_name (info), exception);
       arg_info = g_callable_info_get_arg ((GICallableInfo *) info, i);
       dir = g_arg_info_get_direction (arg_info);
       type_info = g_arg_info_get_type (arg_info);
@@ -371,15 +478,16 @@ seed_gobject_method_invoked (JSContextRef ctx,
 	{
 
 	  if (!seed_gi_make_argument (ctx, arguments[i],
-				      type_info,
+				      type_info, arg_info,
 				      &in_args[n_in_args++], exception))
 	    {
-	      seed_make_exception (ctx, exception, 
-				   "ArgumentError", 
-				   "Unable to make argument %d for" 
+	      seed_make_exception (ctx, exception,
+				   "ArgumentError",
+				   "Unable to make argument %d for"
 				   " function: %s. \n",
 				   i + 1,
-				   g_base_info_get_name ((GIBaseInfo *) info));
+				   g_base_info_get_name ((GIBaseInfo *)
+							 info));
 
 	      g_base_info_unref ((GIBaseInfo *) type_info);
 	      g_base_info_unref ((GIBaseInfo *) arg_info);
@@ -468,6 +576,8 @@ seed_gobject_method_invoked (JSContextRef ctx,
 
       return JSValueMakeNull (ctx);
     }
+
+  in_args_pos = out_args_pos = 0;
   for (i = 0; (i < n_args); i++)
     {
       JSValueRef jsout_val;
@@ -479,14 +589,17 @@ seed_gobject_method_invoked (JSContextRef ctx,
 	{
 	  seed_gi_release_in_arg (g_arg_info_get_ownership_transfer
 				  (arg_info), type_info,
-				  &in_args[i + (instance_method ? 1 : 0)]);
+				  &in_args[in_args_pos +
+					   (instance_method ? 1 : 0)]);
+	  in_args_pos++;
 
 	  g_base_info_unref ((GIBaseInfo *) type_info);
 	  g_base_info_unref ((GIBaseInfo *) arg_info);
 	  continue;
 	}
-      jsout_val = seed_gi_argument_make_js (ctx, &out_values[i],
+      jsout_val = seed_gi_argument_make_js (ctx, &out_args[out_args_pos],
 					    type_info, exception);
+      out_args_pos++;
       if (!JSValueIsNull (ctx, arguments[i]) &&
 	  JSValueIsObject (ctx, arguments[i]))
 	{
@@ -540,8 +653,10 @@ seed_gobject_define_property_from_function_info (JSContextRef ctx,
   method_ref = JSObjectMake (ctx, gobject_method_class,
 			     g_base_info_ref ((GIBaseInfo *) info));
 
+  JSObjectSetPrototype (ctx, method_ref, function_proto);
+
   name = g_base_info_get_name ((GIBaseInfo *) info);
-  if (!strcmp (name, "new"))
+  if (!g_strcmp0 (name, "new"))
     name = "c_new";
   seed_object_set_property (ctx, object, name, method_ref);
   seed_object_set_property (ctx, method_ref, "info",
@@ -558,14 +673,13 @@ seed_gobject_add_methods_for_interfaces (JSContextRef ctx,
 					 JSObjectRef object)
 {
   GIInterfaceInfo *interface;
-  gint n_interfaces, i;
+  GIFunctionInfo *function;
+  gint n_interfaces, i, n_functions, k;
 
   n_interfaces = g_object_info_get_n_interfaces (oinfo);
 
   for (i = 0; i < n_interfaces; i++)
     {
-      GIFunctionInfo *function;
-      gint n_functions, k;
       interface = g_object_info_get_interface (oinfo, i);
 
       n_functions = g_interface_info_get_n_methods (interface);
@@ -654,8 +768,8 @@ seed_gobject_get_class_for_gtype (JSContextRef ctx, GType type)
       GType *interfaces;
       GIFunctionInfo *function;
       GIBaseInfo *interface;
-      gint n_functions, k, i;
-      guint n;
+      gint n_functions, k;
+      guint i, n;
 
       interfaces = g_type_interfaces (type, &n);
       for (i = 0; i < n; i++)
@@ -718,30 +832,6 @@ seed_gobject_finalize (JSObjectRef object)
   g_object_run_dispose (gobject);
 }
 
-static void
-seed_gobject_initialize (JSContextRef ctx, JSObjectRef object)
-{
-  GObject *gobject;
-  GIBaseInfo *base;
-
-  gobject = seed_value_to_object (ctx, (JSValueRef) object, 0);
-  if (!gobject)
-    return;
-
-  base = g_irepository_find_by_gtype (g_irepository_get_default (),
-				      G_OBJECT_TYPE (gobject));
-
-  
-  seed_add_signals_to_object (ctx,
-			      object,
-			      gobject);
-
-  if (!base)
-    return;
-
-  g_assert (g_base_info_get_type (base) == GI_INFO_TYPE_OBJECT);
-}
-
 static JSValueRef
 seed_gobject_get_property (JSContextRef context,
 			   JSObjectRef object,
@@ -753,7 +843,8 @@ seed_gobject_get_property (JSContextRef context,
   char *cproperty_name;
   gint length;
   JSValueRef ret;
-  gint i, len;
+  guint i;
+  gsize len;
 
   b = seed_value_to_object (context, (JSValueRef) object, exception);
   if (!b)
@@ -768,9 +859,8 @@ seed_gobject_get_property (JSContextRef context,
 
   if (!spec)
     {
-
-      len = strlen (cproperty_name);
-      for (i = 0; i < len - 1; i++)
+      len = strlen (cproperty_name) - 1;
+      for (i = 0; i < len; i++)
 	{
 	  if (cproperty_name[i] == '_')
 	    cproperty_name[i] = '-';
@@ -781,13 +871,13 @@ seed_gobject_get_property (JSContextRef context,
 	goto found;
       else
 	{
-	  GIFieldInfo *field = 0;
+	  GIFieldInfo *field = NULL;
 	  GIBaseInfo *info = (GIBaseInfo *)
 	    g_irepository_find_by_gtype (0, G_OBJECT_TYPE (b));
 	  gint n;
 	  const gchar *name;
 
-	  for (i = 0; i < len - 1; i++)
+	  for (i = 0; i < len; i++)
 	    {
 	      if (cproperty_name[i] == '-')
 		cproperty_name[i] = '_';
@@ -795,7 +885,7 @@ seed_gobject_get_property (JSContextRef context,
 
 	  if (!info)
 	    {
-	      return 0;
+	      return NULL;
 	    }
 
 	  n = g_object_info_get_n_fields ((GIObjectInfo *) info);
@@ -804,7 +894,7 @@ seed_gobject_get_property (JSContextRef context,
 	      field = g_object_info_get_field ((GIObjectInfo *) info, i);
 	      name = g_base_info_get_name ((GIBaseInfo *) field);
 
-	      if (!strcmp (name, cproperty_name))
+	      if (!g_strcmp0 (name, cproperty_name))
 		goto found_field;
 	      else
 		{
@@ -821,7 +911,7 @@ seed_gobject_get_property (JSContextRef context,
 	    }
 	  g_base_info_unref ((GIBaseInfo *) info);
 	}
-      return 0;
+      return NULL;
     }
 found:
 
@@ -844,11 +934,13 @@ seed_gobject_set_property (JSContextRef context,
   GValue gval = { 0 };
   GType type;
   gchar *cproperty_name;
-  gint length;
+  gsize length;
+  gsize i, len;
 
-  if (JSValueIsNull (context, value))
+  if (seed_next_gobject_wrapper || JSValueIsNull (context, value))
     return 0;
 
+
   obj = seed_value_to_object (context, object, 0);
 
   length = JSStringGetMaximumUTF8CStringSize (property_name);
@@ -860,7 +952,6 @@ seed_gobject_set_property (JSContextRef context,
 
   if (!spec)
     {
-      gint i, len;
       len = strlen (cproperty_name);
       for (i = 0; i < len; i++)
 	{
@@ -871,7 +962,7 @@ seed_gobject_set_property (JSContextRef context,
 					   cproperty_name);
       if (!spec)
 	{
-	  return 0;
+	  return FALSE;
 	}
     }
 
@@ -883,7 +974,7 @@ seed_gobject_set_property (JSContextRef context,
   seed_gvalue_from_seed_value (context, value, type, &gval, exception);
   if (*exception)
     {
-      return 0;
+      return FALSE;
     }
 
   if (glib_message)
@@ -894,7 +985,7 @@ seed_gobject_set_property (JSContextRef context,
   g_object_set_property (obj, cproperty_name, &gval);
   if (glib_message != 0)
     {
-      seed_make_exception (context, exception, "PropertyError", 
+      seed_make_exception (context, exception, "PropertyError",
 			   glib_message, NULL);
 
       return FALSE;
@@ -905,80 +996,34 @@ seed_gobject_set_property (JSContextRef context,
   return TRUE;
 }
 
-static JSValueRef
-seed_gi_import_namespace (JSContextRef ctx,
-			  JSObjectRef function,
-			  JSObjectRef this_object,
-			  size_t argumentCount,
-			  const JSValueRef arguments[],
-			  JSValueRef * exception)
-{
-  gchar *namespace;
-  const gchar *version = 0;
-  gchar *jsextension;
-  JSStringRef extension_script;
-
-  ctx = eng->context;
-
-  if (argumentCount == 0)
-    {
-      seed_make_exception (ctx, exception,
-			   "ArgumentError",
-			   "Seed.import_namespace"
-			   " expected 1 or 2 arguments, got 0");
-      return JSValueMakeNull (ctx);
-    }
-
-  namespace = seed_value_to_string (ctx, arguments[0], exception);
-  if (argumentCount == 2)
-    {
-      version = seed_value_to_string (ctx, arguments[1], exception);
-    }
-
-  if (version)
-    jsextension = g_strdup_printf("imports.gi.versions.%s = %s; %s = imports.gi.%s;",
-				  namespace, version, namespace, namespace);
-  else
-    jsextension =
-      g_strdup_printf("%s = imports.gi.%s;", namespace, namespace);
-  extension_script = JSStringCreateWithUTF8CString (jsextension);
-  JSEvaluateScript (ctx, extension_script, NULL, NULL, 0, NULL);
-  JSStringRelease (extension_script);
-
-  g_free ((gchar *) namespace);
-  g_free (jsextension);
-
-  return JSValueMakeNull (ctx);
-}
-
 static JSValueRef
 seed_gobject_constructor_convert_to_type (JSContextRef ctx,
 					  JSObjectRef object,
-					  JSType type,
-					  JSValueRef *exception)
+					  JSType type, JSValueRef * exception)
 {
   GType gtype;
-  
+  gchar *as_string;
+
   if (type == kJSTypeString)
     {
       JSValueRef ret;
-      gchar *as_string;
       gtype = (GType) JSObjectGetPrivate (object);
-      
-      as_string = g_strdup_printf("[gobject_constructor %s]", g_type_name (gtype));
+
+      as_string =
+	g_strdup_printf ("[gobject_constructor %s]", g_type_name (gtype));
       ret = seed_value_from_string (ctx, as_string, exception);
       g_free (as_string);
-      
+
       return ret;
     }
   return FALSE;
 }
 
 JSStaticFunction gobject_static_funcs[] = {
-  {"equals", seed_gobject_equals, 0}
-  ,
   {"__debug_ref_count", seed_gobject_ref_count, 0}
   ,
+  {"__property_type", seed_gobject_property_type, 0}
+  ,
   {0, 0, 0}
 };
 
@@ -989,7 +1034,7 @@ JSClassDefinition gobject_def = {
   NULL,				/* Parent Class */
   NULL,				/* Static Values */
   gobject_static_funcs,		/* Static Functions */
-  seed_gobject_initialize,	/* Initialize */
+  NULL,
   seed_gobject_finalize,	/* Finalize */
   NULL,				/* Has Property */
   seed_gobject_get_property,	/* Get Property */
@@ -1078,7 +1123,7 @@ JSClassDefinition gobject_constructor_def = {
   NULL,				/* Get Property Names */
   NULL,				/* Call As Function */
   seed_gobject_constructor_invoked,	/* Call As Constructor */
-  NULL,				/* Has Instance */
+  seed_gobject_has_instance,				/* Has Instance */
   seed_gobject_constructor_convert_to_type
 };
 
@@ -1098,7 +1143,7 @@ JSClassDefinition gobject_named_constructor_def = {
   NULL,				/* Get Property Names */
   NULL,				/* Call As Function */
   seed_gobject_named_constructor_invoked,	/* Call As Constructor */
-  NULL,				/* Has Instance */
+  seed_gobject_has_instance,				/* Has Instance */
   seed_gobject_constructor_convert_to_type
 };
 
@@ -1122,11 +1167,21 @@ JSClassDefinition struct_constructor_def = {
   NULL				/* Convert To Type */
 };
 
+/**
+ * seed_create_function:
+ * @ctx: A valid #SeedContext
+ * @name: The name of the function (used in exceptions).
+ * @func: A #SeedFunctionCallback to implement the function.
+ * @obj: The #SeedObject on which to put the function. 
+ *
+ * Creates a JavaScript object representing a first-class function; when
+ * the function is called from JavaScript, @func will be called. Places
+ * the created function as the property @name on @obj.
+ *
+ */
 void
 seed_create_function (JSContextRef ctx,
-		      gchar * name, 
-		      gpointer func, 
-		      JSObjectRef obj)
+		      gchar * name, gpointer func, JSObjectRef obj)
 {
   JSObjectRef oref;
 
@@ -1140,7 +1195,7 @@ seed_repl_expose (JSContextRef ctx, ...)
   va_list argp;
   void *expose;
   JSObjectRef arrayObj;
-  int i = 0;
+  guint i = 0;
   JSStringRef script;
   JSObjectRef seed = (JSObjectRef) seed_object_get_property (ctx,
 							     JSContextGetGlobalObject
@@ -1154,7 +1209,7 @@ seed_repl_expose (JSContextRef ctx, ...)
 
   while ((expose = va_arg (argp, void *)))
     {
-      g_print ("  Seed.debug_argv[%d] = %p\n", i, expose);
+      g_print ("  Seed.debug_argv[%u] = %p\n", i, expose);
       JSObjectSetPropertyAtIndex (ctx, arrayObj, i++, expose, NULL);
     }
 
@@ -1163,9 +1218,9 @@ seed_repl_expose (JSContextRef ctx, ...)
   seed_object_set_property (ctx, seed, "debug_argv", arrayObj);
 
   script = JSStringCreateWithUTF8CString ("readline = imports.readline;"
-					  "while(1) { try { Seed.print(eval("
+					  "while(1) { try { print(eval("
 					  "readline.readline(\"> \"))); } catch(e) {"
-					  "Seed.print(e.name + \" \" + e.message);}}");
+					  "print(e.name + \" \" + e.message);}}");
 
   JSEvaluateScript (ctx, script, NULL, NULL, 0, NULL);
 
@@ -1212,6 +1267,8 @@ static GOptionEntry seed_args[] = {
   {"seed-no-debug", 0, 0, G_OPTION_ARG_CALLBACK, seed_arg_no_debug_cb,
    "Disable Seed debugging", "FLAGS"},
 #endif /* SEED_ENABLE_DEBUG */
+  {"seed-version", 0, 0, G_OPTION_ARG_NONE, &seed_arg_print_version,
+   "Print libseed version", 0},
   {NULL,},
 };
 
@@ -1232,6 +1289,7 @@ seed_parse_args (int *argc, char ***argv)
 {
   GOptionContext *option_context;
   GOptionGroup *seed_group;
+
   GError *error = NULL;
   gboolean ret = TRUE;
 
@@ -1242,7 +1300,7 @@ seed_parse_args (int *argc, char ***argv)
   /* Initiate any command line options from the backend */
 
   seed_group = seed_get_option_group ();
-  g_option_context_set_main_group (option_context, seed_group);
+  g_option_context_add_group (option_context, seed_group);
 
   if (!g_option_context_parse (option_context, argc, argv, &error))
     {
@@ -1260,114 +1318,49 @@ seed_parse_args (int *argc, char ***argv)
   return ret;
 }
 
+void
+seed_engine_destroy (SeedEngine *eng)
+{
+  JSValueUnprotect (eng->context, eng->global);
+  JSGlobalContextRelease (eng->context);
+  JSContextGroupRelease (eng->group);
+  
+  g_free (eng);
+}
+
 /**
- * seed_init:
+ * seed_init_with_context_group:
  * @argc: A reference to the number of arguments remaining to parse.
  * @argv: A reference to an array of string arguments remaining to parse.
+ * @group: A #SeedContextGroup within which to create the initial context.
  *
- * Initializes a new #SeedEngine.
+ * Initializes a new #SeedEngine. This involves initializing GLib, creating
+ * an initial context (in #group) with all of the default globals, and
+ * initializing various internal parts of Seed.
+ *
+ * This function should only be called once within a single Seed application.
  *
  * Return value: The newly created and initialized #SeedEngine.
  *
  */
 SeedEngine *
-seed_init (gint * argc, gchar *** argv)
+seed_init_with_context_group (gint * argc,
+			      gchar *** argv, JSContextGroupRef group)
 {
-  JSStringRef defaults_script;
 
   g_type_init ();
-  g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_WARNING, seed_log_handler,
-		     0);
+  g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_WARNING, seed_log_handler, 0);
 
   if ((argc != 0) && seed_parse_args (argc, argv) == FALSE)
     {
       SEED_NOTE (MISC, "failed to parse arguments.");
-      return false;
+      return FALSE;
     }
-
-  qname = g_quark_from_static_string ("js-type");
-  qprototype = g_quark_from_static_string ("js-prototype");
-  js_ref_quark = g_quark_from_static_string ("js-ref");
-
-  eng = (SeedEngine *) g_malloc (sizeof (SeedEngine));
-
-  context_group = JSContextGroupCreate ();
-
-  eng->context = JSGlobalContextCreateInGroup (context_group, NULL);
-  eng->global = JSContextGetGlobalObject (eng->context);
-  eng->group = context_group;
-  eng->search_path = NULL;
-
-  gobject_class = JSClassCreate (&gobject_def);
-  JSClassRetain (gobject_class);
-  gobject_method_class = JSClassCreate (&gobject_method_def);
-  JSClassRetain (gobject_method_class);
-  gobject_constructor_class = JSClassCreate (&gobject_constructor_def);
-  JSClassRetain (gobject_constructor_class);
-  gobject_named_constructor_class =
-    JSClassCreate (&gobject_named_constructor_def);
-  JSClassRetain (gobject_named_constructor_class);
-  gobject_signal_class = JSClassCreate (seed_get_signal_class ());
-  JSClassRetain (gobject_signal_class);
-  seed_callback_class = JSClassCreate (&seed_callback_def);
-  JSClassRetain (seed_callback_class);
-  seed_struct_constructor_class = JSClassCreate (&struct_constructor_def);
-  JSClassRetain (seed_struct_constructor_class);
-  gobject_init_method_class = JSClassCreate (&gobject_init_method_def);
-  JSClassRetain (gobject_init_method_class);
-
-  g_type_set_qdata (G_TYPE_OBJECT, qname, gobject_class);
-
-  seed_obj_ref = JSObjectMake (eng->context, NULL, NULL);
-  seed_object_set_property (eng->context, eng->global, "Seed", seed_obj_ref);
-  JSValueProtect (eng->context, seed_obj_ref);
-
-  seed_create_function (eng->context, "import_namespace",
-			&seed_gi_import_namespace, seed_obj_ref);
-
-  g_irepository_require (g_irepository_get_default (), "GObject", NULL, 0, 0);
-  g_irepository_require (g_irepository_get_default (), "GIRepository",
-			 NULL, 0, 0);
   
-  seed_initialize_importer (eng->context, eng->global);
-
-  seed_init_builtins (eng, argc, argv);
-  seed_closures_init ();
-  seed_structs_init ();
-
-  seed_gtype_init (eng);
-
-  defaults_script =
-	  JSStringCreateWithUTF8CString ("try{Seed.include(\"/usr/share/"
-					 "seed/extensions/Seed.js\");} catch(e){}"
-					 "Seed.include(\"/usr/local/share"
-					 "/seed/extensions/Seed.js\");");
-  JSEvaluateScript (eng->context, defaults_script, NULL, NULL, 0, NULL);
-  JSStringRelease (defaults_script);
-
-  base_info_info =
-    g_irepository_find_by_name (0, "GIRepository", "IBaseInfo");
-  
-  return eng;
-}
-
-
-
-SeedEngine *
-seed_init_with_context_group (gint * argc, 
-			      gchar *** argv,
-			      JSContextGroupRef group)
-{
-  JSStringRef defaults_script;
-
-  g_type_init ();
-  g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_WARNING, seed_log_handler,
-		     0);
-
-  if ((argc != 0) && seed_parse_args (argc, argv) == FALSE)
+  if (seed_arg_print_version)
     {
-      SEED_NOTE (MISC, "failed to parse arguments.");
-      return false;
+      g_print("%s\n", "Seed " VERSION);
+      exit(EXIT_SUCCESS);
     }
 
   qname = g_quark_from_static_string ("js-type");
@@ -1383,6 +1376,9 @@ seed_init_with_context_group (gint * argc,
   eng->group = context_group;
   eng->search_path = NULL;
 
+  function_proto = (JSObjectRef)
+    seed_simple_evaluate (eng->context, "Function.prototype", NULL);
+
   gobject_class = JSClassCreate (&gobject_def);
   JSClassRetain (gobject_class);
   gobject_method_class = JSClassCreate (&gobject_method_def);
@@ -1407,9 +1403,6 @@ seed_init_with_context_group (gint * argc,
   seed_object_set_property (eng->context, eng->global, "Seed", seed_obj_ref);
   JSValueProtect (eng->context, seed_obj_ref);
 
-  seed_create_function (eng->context, "import_namespace",
-			&seed_gi_import_namespace, seed_obj_ref);
-
   g_irepository_require (g_irepository_get_default (), "GObject", NULL, 0, 0);
   g_irepository_require (g_irepository_get_default (), "GIRepository",
 			 NULL, 0, 0);
@@ -1422,16 +1415,37 @@ seed_init_with_context_group (gint * argc,
 
   seed_gtype_init (eng);
 
+
   defaults_script =
-    JSStringCreateWithUTF8CString ("try{Seed.include(\"/usr/share/"
-				   "seed/extensions/Seed.js\");} catch(e){}"
-				   "Seed.include(\"/usr/local/share"
-				   "/seed/extensions/Seed.js\");");
+    JSStringCreateWithUTF8CString ("Seed.include(\"" SEED_PREFIX_PATH
+				   "extensions/Seed.js\");");
+
   JSEvaluateScript (eng->context, defaults_script, NULL, NULL, 0, NULL);
-  JSStringRelease (defaults_script);
 
   base_info_info =
     g_irepository_find_by_name (0, "GIRepository", "IBaseInfo");
-  
+
   return eng;
 }
+
+/**
+ * seed_init:
+ * @argc: A reference to the number of arguments remaining to parse.
+ * @argv: A reference to an array of string arguments remaining to parse.
+ *
+ * Initializes a new #SeedEngine. This involves initializing GLib, creating
+ * an initial context with all of the default globals, and initializing
+ * various internal parts of Seed.
+ *
+ * This function should only be called once within a single Seed application.
+ *
+ * Return value: The newly created and initialized #SeedEngine.
+ *
+ */
+SeedEngine *
+seed_init (gint * argc, gchar *** argv)
+{
+  context_group = JSContextGroupCreate ();
+
+  return seed_init_with_context_group (argc, argv, context_group);
+}
